import tools.frendy as fdy
import tools.depletion_chain as dc
import h5py


####################################################################################
##                         MODIFY LINES IN THIS BLOCK
base = "/path/to/ENDF-B-VIII.0/neutrons/"
tslbase = "/path/to/ENDF-B-VIII.0/thermal_scatt/"
lib_name = "ENDF/B-VIII.0"

temps = [293.0, 500.0, 600.0, 800.0, 1000.0, 1500.0, 2000.0]

hdf5_file_name = "endf8_shem281.h5"

# Set the default group strucutre
fdy.set_default_group_structure("SHEM-281")
fdy.set_default_max_legendre_moments(3)

####################################################################################


# First we initialize the file with the depletion chain data.
# Must be done first as the save method will delete file first if it already exists !
dep_chain = dc.build_depletion_chain("chain_casl_pwr.xml")
dep_chain.save(hdf5_file_name)

# Initialize HDF5 with the library information
print(fdy.get_default_group_structure())
h5 = h5py.File(hdf5_file_name, "r+")
h5.attrs["group-structure"] = fdy.get_default_group_structure().name
h5.attrs["group-bounds"] = fdy.get_default_group_structure().bounds
h5.attrs["ngroups"] = fdy.get_default_group_structure().ngroups
h5.attrs["condensation-scheme"] = fdy.get_default_group_structure().condensation_scheme
h5.attrs["cmfd-condensation-scheme"] = (
    fdy.get_default_group_structure().cmfd_condensation_scheme
)
h5.attrs["first-resonance-group"] = fdy.get_default_group_structure().first_res_grp
h5.attrs["last-resonance-group"] = fdy.get_default_group_structure().last_res_grp
h5.attrs["library"] = lib_name

# We process U235 first so that we can steal its fission spectrum for
# performing transport correciton calculations
N = fdy.FrendyMG()
N.name = "U235"
N.endf_file = base + "n-092_U_235.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

chi = N.chi[:]

# Process TSL based evaluations
N = fdy.FrendyMG()
N.name = "H1_H2O"
N.endf_file = base + "n-001_H_001.endf"
N.tsl_file = tslbase + "tsl-HinH2O.endf"
N.tsl_type = "hh2o"
N.label = "H1 in H2O from ENDF/B-8.0"
N.temps = [
    283.6,
    293.6,
    300.0,
    323.6,
    350.0,
    373.6,
    400.0,
    423.6,
    450.0,
    473.6,
    500.0,
    523.6,
    550.0,
    573.6,
    600.0,
    623.6,
    650.0,
    800.0,
]
N.dilutions = [1.0e10]
N.process(h5, chi)  # Processed with in-scatter transport correction !

N = fdy.FrendyMG()
N.name = "H2_D2O"
N.endf_file = base + "n-001_H_002.endf"
N.tsl_file = tslbase + "tsl-DinD2O.endf"
N.tsl_type = "dd2o"
N.label = "H2 in D2O from ENDF/B-8.0"
N.temps = [
    283.6,
    293.6,
    300.0,
    323.6,
    350.0,
    373.6,
    400.0,
    423.6,
    450.0,
    473.6,
    500.0,
    523.6,
    550.0,
    573.6,
    600.0,
    623.6,
    650.0,
]
N.dilutions = [1.0e10]
N.process(h5, chi)  # Processed with in-scatter transport correction !

# Free Gas Nuclides
N = fdy.FrendyMG()
N.name = "H1"
N.endf_file = base + "n-001_H_001.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "H2"
N.endf_file = base + "n-001_H_002.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "He3"
N.endf_file = base + "n-002_He_003.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "He4"
N.endf_file = base + "n-002_He_004.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Li6"
N.endf_file = base + "n-003_Li_006.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Li7"
N.endf_file = base + "n-003_Li_007.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Be9"
N.endf_file = base + "n-004_Be_009.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "B10"
N.endf_file = base + "n-005_B_010.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "B11"
N.endf_file = base + "n-005_B_011.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "C12"
N.endf_file = base + "n-006_C_012.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "C13"
N.endf_file = base + "n-006_C_013.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "N14"
N.endf_file = base + "n-007_N_014.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "N15"
N.endf_file = base + "n-007_N_015.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "O16"
N.endf_file = base + "n-008_O_016.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "O17"
N.endf_file = base + "n-008_O_017.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "O18"
N.endf_file = base + "n-008_O_018.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Na23"
N.endf_file = base + "n-011_Na_023.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mg24"
N.endf_file = base + "n-012_Mg_024.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mg25"
N.endf_file = base + "n-012_Mg_025.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mg26"
N.endf_file = base + "n-012_Mg_026.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Al27"
N.endf_file = base + "n-013_Al_027.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Si28"
N.endf_file = base + "n-014_Si_028.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Si29"
N.endf_file = base + "n-014_Si_029.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Si30"
N.endf_file = base + "n-014_Si_030.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ar36"
N.endf_file = base + "n-018_Ar_036.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ar38"
N.endf_file = base + "n-018_Ar_038.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ar40"
N.endf_file = base + "n-018_Ar_040.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cr50"
N.endf_file = base + "n-024_Cr_050.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cr52"
N.endf_file = base + "n-024_Cr_052.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cr53"
N.endf_file = base + "n-024_Cr_053.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cr54"
N.endf_file = base + "n-024_Cr_054.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mn55"
N.endf_file = base + "n-025_Mn_055.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Fe54"
N.endf_file = base + "n-026_Fe_054.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Fe55"
N.endf_file = base + "n-026_Fe_055.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Fe56"
N.endf_file = base + "n-026_Fe_056.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Fe57"
N.endf_file = base + "n-026_Fe_057.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Fe58"
N.endf_file = base + "n-026_Fe_058.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Co59"
N.endf_file = base + "n-027_Co_059.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ni58"
N.endf_file = base + "n-028_Ni_058.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ni60"
N.endf_file = base + "n-028_Ni_060.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ni61"
N.endf_file = base + "n-028_Ni_061.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ni62"
N.endf_file = base + "n-028_Ni_062.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ni64"
N.endf_file = base + "n-028_Ni_064.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cu63"
N.endf_file = base + "n-029_Cu_063.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cu65"
N.endf_file = base + "n-029_Cu_065.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Br81"
N.endf_file = base + "n-035_Br_081.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Kr82"
N.endf_file = base + "n-036_Kr_082.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Kr83"
N.endf_file = base + "n-036_Kr_083.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Kr84"
N.endf_file = base + "n-036_Kr_084.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Kr85"
N.endf_file = base + "n-036_Kr_085.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Kr86"
N.endf_file = base + "n-036_Kr_086.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sr89"
N.endf_file = base + "n-038_Sr_089.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sr90"
N.endf_file = base + "n-038_Sr_090.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Y89"
N.endf_file = base + "n-039_Y_089.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Y90"
N.endf_file = base + "n-039_Y_090.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Y91"
N.endf_file = base + "n-039_Y_091.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr90"
N.endf_file = base + "n-040_Zr_090.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr91"
N.endf_file = base + "n-040_Zr_091.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr92"
N.endf_file = base + "n-040_Zr_092.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr93"
N.endf_file = base + "n-040_Zr_093.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr94"
N.endf_file = base + "n-040_Zr_094.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr95"
N.endf_file = base + "n-040_Zr_095.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Zr96"
N.endf_file = base + "n-040_Zr_096.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nb95"
N.endf_file = base + "n-041_Nb_095.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo92"
N.endf_file = base + "n-042_Mo_092.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo94"
N.endf_file = base + "n-042_Mo_094.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo95"
N.endf_file = base + "n-042_Mo_095.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo96"
N.endf_file = base + "n-042_Mo_096.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo97"
N.endf_file = base + "n-042_Mo_097.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo98"
N.endf_file = base + "n-042_Mo_098.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo99"
N.endf_file = base + "n-042_Mo_099.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Mo100"
N.endf_file = base + "n-042_Mo_100.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tc99"
N.endf_file = base + "n-043_Tc_099.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru99"
N.endf_file = base + "n-044_Ru_099.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru100"
N.endf_file = base + "n-044_Ru_100.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru101"
N.endf_file = base + "n-044_Ru_101.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru102"
N.endf_file = base + "n-044_Ru_102.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru103"
N.endf_file = base + "n-044_Ru_103.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru104"
N.endf_file = base + "n-044_Ru_104.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru105"
N.endf_file = base + "n-044_Ru_105.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ru106"
N.endf_file = base + "n-044_Ru_106.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Rh103"
N.endf_file = base + "n-045_Rh_103.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Rh105"
N.endf_file = base + "n-045_Rh_105.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pd104"
N.endf_file = base + "n-046_Pd_104.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pd105"
N.endf_file = base + "n-046_Pd_105.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pd106"
N.endf_file = base + "n-046_Pd_106.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pd107"
N.endf_file = base + "n-046_Pd_107.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pd108"
N.endf_file = base + "n-046_Pd_108.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ag107"
N.endf_file = base + "n-047_Ag_107.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ag109"
N.endf_file = base + "n-047_Ag_109.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ag110m1"
N.endf_file = base + "n-047_Ag_110m1.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ag111"
N.endf_file = base + "n-047_Ag_111.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd106"
N.endf_file = base + "n-048_Cd_106.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd108"
N.endf_file = base + "n-048_Cd_108.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd110"
N.endf_file = base + "n-048_Cd_110.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd111"
N.endf_file = base + "n-048_Cd_111.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd112"
N.endf_file = base + "n-048_Cd_112.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd113"
N.endf_file = base + "n-048_Cd_113.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd114"
N.endf_file = base + "n-048_Cd_114.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cd116"
N.endf_file = base + "n-048_Cd_116.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "In113"
N.endf_file = base + "n-049_In_113.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "In115"
N.endf_file = base + "n-049_In_115.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn112"
N.endf_file = base + "n-050_Sn_112.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn114"
N.endf_file = base + "n-050_Sn_114.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn115"
N.endf_file = base + "n-050_Sn_115.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn116"
N.endf_file = base + "n-050_Sn_116.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn117"
N.endf_file = base + "n-050_Sn_117.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn118"
N.endf_file = base + "n-050_Sn_118.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn119"
N.endf_file = base + "n-050_Sn_119.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn120"
N.endf_file = base + "n-050_Sn_120.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn122"
N.endf_file = base + "n-050_Sn_122.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sn124"
N.endf_file = base + "n-050_Sn_124.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sb121"
N.endf_file = base + "n-051_Sb_121.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sb123"
N.endf_file = base + "n-051_Sb_123.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sb125"
N.endf_file = base + "n-051_Sb_125.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Te127m1"
N.endf_file = base + "n-052_Te_127m1.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Te129m1"
N.endf_file = base + "n-052_Te_129m1.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Te132"
N.endf_file = base + "n-052_Te_132.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "I127"
N.endf_file = base + "n-053_I_127.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "I129"
N.endf_file = base + "n-053_I_129.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "I131"
N.endf_file = base + "n-053_I_131.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "I135"
N.endf_file = base + "n-053_I_135.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe128"
N.endf_file = base + "n-054_Xe_128.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe129"
N.endf_file = base + "n-054_Xe_129.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe130"
N.endf_file = base + "n-054_Xe_130.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe131"
N.endf_file = base + "n-054_Xe_131.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe132"
N.endf_file = base + "n-054_Xe_132.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe133"
N.endf_file = base + "n-054_Xe_133.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe134"
N.endf_file = base + "n-054_Xe_134.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe135"
N.endf_file = base + "n-054_Xe_135.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Xe136"
N.endf_file = base + "n-054_Xe_136.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cs133"
N.endf_file = base + "n-055_Cs_133.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cs134"
N.endf_file = base + "n-055_Cs_134.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cs135"
N.endf_file = base + "n-055_Cs_135.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cs136"
N.endf_file = base + "n-055_Cs_136.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cs137"
N.endf_file = base + "n-055_Cs_137.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ba134"
N.endf_file = base + "n-056_Ba_134.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ba137"
N.endf_file = base + "n-056_Ba_137.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ba140"
N.endf_file = base + "n-056_Ba_140.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "La139"
N.endf_file = base + "n-057_La_139.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "La140"
N.endf_file = base + "n-057_La_140.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ce140"
N.endf_file = base + "n-058_Ce_140.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ce141"
N.endf_file = base + "n-058_Ce_141.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ce142"
N.endf_file = base + "n-058_Ce_142.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ce143"
N.endf_file = base + "n-058_Ce_143.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ce144"
N.endf_file = base + "n-058_Ce_144.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pr141"
N.endf_file = base + "n-059_Pr_141.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pr143"
N.endf_file = base + "n-059_Pr_143.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd142"
N.endf_file = base + "n-060_Nd_142.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd143"
N.endf_file = base + "n-060_Nd_143.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd144"
N.endf_file = base + "n-060_Nd_144.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd145"
N.endf_file = base + "n-060_Nd_145.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd146"
N.endf_file = base + "n-060_Nd_146.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd147"
N.endf_file = base + "n-060_Nd_147.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd148"
N.endf_file = base + "n-060_Nd_148.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Nd150"
N.endf_file = base + "n-060_Nd_150.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pm147"
N.endf_file = base + "n-061_Pm_147.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pm148"
N.endf_file = base + "n-061_Pm_148.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pm148m1"
N.endf_file = base + "n-061_Pm_148m1.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pm149"
N.endf_file = base + "n-061_Pm_149.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pm151"
N.endf_file = base + "n-061_Pm_151.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm147"
N.endf_file = base + "n-062_Sm_147.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm148"
N.endf_file = base + "n-062_Sm_148.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm149"
N.endf_file = base + "n-062_Sm_149.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm150"
N.endf_file = base + "n-062_Sm_150.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm151"
N.endf_file = base + "n-062_Sm_151.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm152"
N.endf_file = base + "n-062_Sm_152.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm153"
N.endf_file = base + "n-062_Sm_153.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Sm154"
N.endf_file = base + "n-062_Sm_154.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu151"
N.endf_file = base + "n-063_Eu_151.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu152"
N.endf_file = base + "n-063_Eu_152.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu153"
N.endf_file = base + "n-063_Eu_153.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu154"
N.endf_file = base + "n-063_Eu_154.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu155"
N.endf_file = base + "n-063_Eu_155.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Eu156"
N.endf_file = base + "n-063_Eu_156.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd152"
N.endf_file = base + "n-064_Gd_152.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd154"
N.endf_file = base + "n-064_Gd_154.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd155"
N.endf_file = base + "n-064_Gd_155.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd156"
N.endf_file = base + "n-064_Gd_156.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd157"
N.endf_file = base + "n-064_Gd_157.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd158"
N.endf_file = base + "n-064_Gd_158.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Gd160"
N.endf_file = base + "n-064_Gd_160.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tb159"
N.endf_file = base + "n-065_Tb_159.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tb160"
N.endf_file = base + "n-065_Tb_160.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tb161"
N.endf_file = base + "n-065_Tb_161.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Dy160"
N.endf_file = base + "n-066_Dy_160.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Dy161"
N.endf_file = base + "n-066_Dy_161.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Dy162"
N.endf_file = base + "n-066_Dy_162.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Dy163"
N.endf_file = base + "n-066_Dy_163.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Dy164"
N.endf_file = base + "n-066_Dy_164.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ho165"
N.endf_file = base + "n-067_Ho_165.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er162"
N.endf_file = base + "n-068_Er_162.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er164"
N.endf_file = base + "n-068_Er_164.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er166"
N.endf_file = base + "n-068_Er_166.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er167"
N.endf_file = base + "n-068_Er_167.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er168"
N.endf_file = base + "n-068_Er_168.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er169"
N.endf_file = base + "n-068_Er_169.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Er170"
N.endf_file = base + "n-068_Er_170.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tm169"
N.endf_file = base + "n-069_Tm_169.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tm170"
N.endf_file = base + "n-069_Tm_170.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Tm171"
N.endf_file = base + "n-069_Tm_171.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf174"
N.endf_file = base + "n-072_Hf_174.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf176"
N.endf_file = base + "n-072_Hf_176.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf177"
N.endf_file = base + "n-072_Hf_177.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf178"
N.endf_file = base + "n-072_Hf_178.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf179"
N.endf_file = base + "n-072_Hf_179.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf180"
N.endf_file = base + "n-072_Hf_180.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Hf181"
N.endf_file = base + "n-072_Hf_181.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ta181"
N.endf_file = base + "n-073_Ta_181.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Ta182"
N.endf_file = base + "n-073_Ta_182.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.dilutions = [1.0e10]  # Fission Product
N.process(h5)

N = fdy.FrendyMG()
N.name = "Th230"
N.endf_file = base + "n-090_Th_230.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Th231"
N.endf_file = base + "n-090_Th_231.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Th232"
N.endf_file = base + "n-090_Th_232.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Th234"
N.endf_file = base + "n-090_Th_234.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pa231"
N.endf_file = base + "n-091_Pa_231.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pa232"
N.endf_file = base + "n-091_Pa_232.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pa233"
N.endf_file = base + "n-091_Pa_233.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U232"
N.endf_file = base + "n-092_U_232.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U233"
N.endf_file = base + "n-092_U_233.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U234"
N.endf_file = base + "n-092_U_234.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U236"
N.endf_file = base + "n-092_U_236.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U237"
N.endf_file = base + "n-092_U_237.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "U238"
N.endf_file = base + "n-092_U_238.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Np236"
N.endf_file = base + "n-093_Np_236.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Np237"
N.endf_file = base + "n-093_Np_237.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Np238"
N.endf_file = base + "n-093_Np_238.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Np239"
N.endf_file = base + "n-093_Np_239.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu236"
N.endf_file = base + "n-094_Pu_236.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu237"
N.endf_file = base + "n-094_Pu_237.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu238"
N.endf_file = base + "n-094_Pu_238.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu239"
N.endf_file = base + "n-094_Pu_239.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu240"
N.endf_file = base + "n-094_Pu_240.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu241"
N.endf_file = base + "n-094_Pu_241.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu242"
N.endf_file = base + "n-094_Pu_242.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu243"
N.endf_file = base + "n-094_Pu_243.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Pu244"
N.endf_file = base + "n-094_Pu_244.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Am241"
N.endf_file = base + "n-095_Am_241.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Am242m1"
N.endf_file = base + "n-095_Am_242m1.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Am243"
N.endf_file = base + "n-095_Am_243.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cm242"
N.endf_file = base + "n-096_Cm_242.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cm243"
N.endf_file = base + "n-096_Cm_243.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cm244"
N.endf_file = base + "n-096_Cm_244.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cm245"
N.endf_file = base + "n-096_Cm_245.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

N = fdy.FrendyMG()
N.name = "Cm246"
N.endf_file = base + "n-096_Cm_246.endf"
N.label = N.name + " from ENDF/B-8.0"
N.temps = temps
N.process(h5)

h5.close()



from .._scarabee import (
    NDLibrary,
    Material,
    CrossSection,
    DiffusionCrossSection,
    DiffusionData,
    ReflectorSN,
    set_logging_level,
    scarabee_log,
    LogLevel,
)
from .nodal_flux import NodalFlux1D

import numpy as np
#import matplotlib.pyplot as plt


class Reflector:
    """
    A Reflector instance is responsible for performing transport calculations
    necessary to produce few-group cross sections for the reflector of an LWR.
    The core baffle cross sections are self-shielded as an infinite slab,
    using the Roman two-term approximation. The calculation is performed using
    a 1D Sn simulation, in the group structure of the nuclear data library.
    This removes the need to obtain a fine-group spectrum that would be used
    to condense to an intermediate group structure.

    Parameters
    ----------
    fuel : CrossSection
        Homogenized cross section which is representative of a pin cell.
        This is typically obtained from a previous lattice calcualtion.
    moderator : CrossSection
        Material cross sections for the moderator at desired temperature
        and density.
    assembly_width : float
        Width of a single fuel assembly (and the reflector to be modeled).
    gap_width : float
        Width of the moderator gap between the assembly and the core baffle.
    baffle_width : float
        Width of the core baffle.
    baffle : Material
        Material for the core baffle at desired temperature and density.
    ndl : NDLibrary
        Nuclear data library for constructing the baffle cross sections.

    Attributes
    ----------
    condensation_scheme : list of pairs of ints
        Defines how the energy groups will be condensed to the few-group
        structure used in nodal calculations.
    fuel : CrossSection
        Cross sections for a homogenized fuel assembly.
    moderator : CrossSection
        Cross sections for the moderator.
    assembly_width : float
        Width of fuel assembly and reflector.
    gap_width : float
        Width of the moderator gap between a fuel assembly and the core baffle.
    baffle_width : float
        Width of the core baffle.
    baffle : CrossSection
        Self-shielded cross sections for the core baffle.
    nangles : int
        Number of angles used in the Sn solver. Default is 16. Must be one of:
        2, 4, 6, 8, 10, 12, 14, 16, 32, 64, 128.
    anisotropic : bool
        If True, the reflector calculation is performed with explicit
        anisotropic scattering. Otherwise, the transport correction with
        isotropic scattering is used. Default value is False.
    keff_tolerance : float
        Convergence criteria for keff. Default is 1.E-5.
    flux_tolerance : float
        Convergence criteria for the flux. Default is 1.E-5.
    diffusion_xs : DiffusionCrossSection
        The few-group diffusion group constants for the reflector.
    adf : ndarray
        The assembly discontinuity factors.
    diffusion_data : DiffusionData
        The few-group diffusion cross sections and ADFs for the reflector.
    """

    def __init__(
        self,
        fuel: CrossSection,
        moderator: CrossSection,
        assembly_width: float,
        gap_width: float,
        baffle_width: float,
        baffle: Material,
        ndl: NDLibrary,
    ):
        self.fuel = fuel
        self.fuel.name = "Fuel"
        self.moderator = moderator
        self.moderator.name = "Moderator"
        self.assembly_width = assembly_width
        self.gap_width = gap_width
        self.baffle_width = baffle_width
        self.condensation_scheme = ndl.condensation_scheme
        self.nangles = 16
        self.anisotropic = False
        self.adf = None
        self.diffusion_xs = None
        self.diffusion_data = None

        # No Dancoff correction, as looking at 1D isolated slab for baffle
        Ee = 1.0 / (2.0 * self.baffle_width)
        self.baffle = baffle.roman_xs(0.0, Ee, ndl)
        self.baffle.name = "Baffle"

        self.keff_tolerance = 1.0e-5
        self.flux_tolerance = 1.0e-5

        if self.gap_width + self.baffle_width >= self.assembly_width:
            raise RuntimeError(
                "The assembly width is smaller than the sum of the gap and baffle widths."
            )

    def solve(self) -> None:
        """
        Runs a 1D annular problem to generate few group cross sections for the
        reflector, with the core baffle.
        """
        scarabee_log(LogLevel.Info, "Starting reflector calculation.")

        if self.condensation_scheme is None:
            raise RuntimeError(
                "Cannot perform reflector calculation without condensation scheme."
            )

        # We start by making a ReflectorSn to do 1D calculation
        dx = []
        mats = []

        # We add 2 fuel assemblies worth of homogenized core
        NF = 2 * 10 * 17
        dr = 2.0 * self.assembly_width / (float(NF))
        dx += [dr] * NF
        mats += [self.fuel] * NF

        # We now add one ring for the gap
        NG = 3
        dr = self.gap_width / (float(NG))
        dx += [dr] * NG
        mats += [self.moderator] * NG

        # Now we add 20 regions for the baffle
        NB = 20
        dr = self.baffle_width / float(NB)
        dx += [dr] * NB
        mats += [self.baffle] * NB

        # Now we add the outer water reflector regions
        ref_width = self.assembly_width - self.gap_width - self.baffle_width
        NR = int(ref_width / 0.02) + 1
        dr = ref_width / float(NR)
        dx += [dr] * NR
        mats += [self.moderator] * NR

        ref_sn = ReflectorSN(mats, dx, self.nangles, self.anisotropic)
        set_logging_level(LogLevel.Warning)
        ref_sn.solve()
        set_logging_level(LogLevel.Info)
        scarabee_log(LogLevel.Info, "")
        scarabee_log(LogLevel.Info, "Kinf: {:.5f}".format(ref_sn.keff))
        scarabee_log(LogLevel.Info, "")
        scarabee_log(LogLevel.Info, "Generating diffusion data.")

        few_group_flux = np.zeros((len(self.condensation_scheme), NF + NG + NB + NR))
        for i in range(NF + 1 + NB + NR):
            for G in range(len(self.condensation_scheme)):
                g_min = self.condensation_scheme[G][0]
                g_max = self.condensation_scheme[G][1]
                for g in range(g_min, g_max + 1):
                    few_group_flux[G, i] += ref_sn.flux(i, g)
        dx = np.array(dx)
        x = np.zeros(len(dx))
        for i in range(len(dx)):
            if i == 0:
                x[0] = 0.5 * dx[0]
            else:
                x[i] = x[i - 1] + 0.5 * (dx[i - 1] + dx[i])

        # Here we compute the cross sections
        ref_homog_xs = ref_sn.homogenize(list(range(NF, NF + NG + NB + NR)))
        ref_homog_spec = ref_sn.homogenize_flux_spectrum(
            list(range(NF, NF + NG + NB + NR))
        )
        ref_homog_diff_xs = ref_homog_xs.diffusion_xs()
        self.diffusion_xs = ref_homog_diff_xs.condense(
            self.condensation_scheme, ref_homog_spec
        )

        fuel_homog_xs = ref_sn.homogenize(list(range(0, NF)))
        fuel_homog_spec = ref_sn.homogenize_flux_spectrum(list(range(0, NF)))
        fuel_homog_diff_xs = fuel_homog_xs.diffusion_xs()
        fuel_diffusion_xs = fuel_homog_diff_xs.condense(
            self.condensation_scheme, fuel_homog_spec
        )

        # Obtain net currents at node boundaries and average flux
        avg_flx_ref = np.zeros(len(self.condensation_scheme))
        avg_flx_fuel = np.zeros(len(self.condensation_scheme))
        j_0 = np.zeros(len(self.condensation_scheme))
        j_mid = np.zeros(len(self.condensation_scheme))
        j_max = np.zeros(len(self.condensation_scheme))
        s_mid = NF
        s_max = ref_sn.nsurfaces - 1
        for g in range(len(self.condensation_scheme)):
            avg_flx_fuel[g] = np.mean(few_group_flux[g, :NF])
            avg_flx_ref[g] = np.mean(few_group_flux[g, NF:])

            g_min = self.condensation_scheme[g][0]
            g_max = self.condensation_scheme[g][1]
            for gg in range(g_min, g_max + 1):
                j_0[g] += ref_sn.current(0, gg)
                j_mid[g] += ref_sn.current(s_mid, gg)
                j_max[g] += ref_sn.current(s_max, gg)

        # Do nodal calculation to obtain homogeneous flux
        x_fuel = np.sum(dx[:NF])
        x_ref_end = np.sum(dx)

        fuel_node = NodalFlux1D(
            0.0, x_fuel, ref_sn.keff, fuel_diffusion_xs, avg_flx_fuel, j_0, j_mid
        )
        ref_node = NodalFlux1D(
            x_fuel, x_ref_end, ref_sn.keff, self.diffusion_xs, avg_flx_ref, j_mid, j_max
        )

        # Plot reference Sn flux and nodal flux
        # for g in range(len(self.condensation_scheme)):
        #    nodal_flux = np.zeros(len(x))
        #    nodal_flux[:NF] = fuel_node(x[:NF], g)
        #    nodal_flux[NF:] = ref_node(x[NF:], g)
        #    plt.plot(x, few_group_flux[g, :], label="Sn")
        #    plt.plot(x, nodal_flux, label="Nodal")
        #    plt.xlabel("x [cm]")
        #    plt.ylabel("Flux [Arb. Units]")
        #    plt.title("Group {:}".format(g))
        #    plt.show()

        # Compute the ADFs
        self.adf = np.zeros((len(self.condensation_scheme), 6))
        for G in range(len(self.condensation_scheme)):
            heter_flx_fuel = few_group_flux[G, NF - 1]
            homog_flx_fuel = fuel_node.pos_surf_flux(G)

            heter_flx_ref = few_group_flux[G, NF]
            homog_flx_ref = ref_node.neg_surf_flux(G)

            f_fuel = heter_flx_fuel / homog_flx_fuel
            f_ref = heter_flx_ref / homog_flx_ref

            # Normalize to fuel DF
            f_ref = f_ref / f_fuel

            self.adf[G, :] = f_ref

        # Create the diffusion data
        self.diffusion_data = DiffusionData(self.diffusion_xs)
        self.diffusion_data.adf = self.adf



from .._scarabee import (
    NDLibrary,
    MaterialComposition,
    Material,
    CrossSection,
    PinCellType,
    SimplePinCell,
    PinCell,
    MOCDriver,
    MixingFraction,
    DepletionChain,
    DepletionMatrix,
    mix_materials,
    build_depletion_matrix,
)
import numpy as np
from typing import Optional, List
import copy


class FuelPin:
    """
    Represents a generic fuel pin for a PWR.

    Parameters
    ----------
    fuel : Material
        Material which describes the fuel composition, density, and temperature.
    fuel_radius : float
        Outer radius of the fuel pellet region.
    gap : Material, optional
        Material which describes the composition, density, and temperature of
        the gap between the fuel pellet and the cladding, if present.
    gap_radius : float, optional
        Outer radius of the gap material, if present.
    clad : Material
        Material which describes the cladding composition, density, and
        temperature.
    clad_radius : float
        Outer radius of the cladding.
    num_fuel_rings : int, default 1
        Number of rings which should be used to discretize the fuel material.
        Each ring will be self-shielded and depleted separately.

    Attributes
    ----------
    fuel_radius : float
        Outer radius of the fuel pellet region.
    fuel_ring_materials : list of list of Material
        Contains the Material in each fuel ring for each depletion time step.
    fuel_ring_flux_spectra : list of list of ndarray
        Contains the average flux spectrum in each fuel ring for each depletion
        time step.
    fuel_dancoff_corrections : list of float
        Dancoff corrections to be used when self-shielding the fuel at each
        depletion time step.
    gap : Material, optional
        Material which describes the composition, density, and temperature of
        the gap between the fuel pellet and the cladding, if present.
    gap_radius : float, optional
        Outer radius of the gap material, if present.
    clad : Material
        Material which describes the cladding composition, density, and
        temperature.
    clad_radius : float
        Outer radius of the cladding.
    clad_dancoff_corrections : list of float
        Dancoff corrections to be used when self-shielding the cladding at each
        depletion time step.
    num_fuel_rings : int, default 1
        Number of rings which should be used to discretize the fuel material.
        Each ring will be self-shielded and depleted separately.
    """

    def __init__(
        self,
        fuel: Material,
        fuel_radius: float,
        clad: Material,
        clad_radius,
        gap: Optional[Material],
        gap_radius: Optional[float],
        num_fuel_rings: int = 1,
    ):
        if fuel_radius <= 0.0:
            raise ValueError("Fuel radius must be > 0.")
        self._fuel_radius = fuel_radius

        if num_fuel_rings <= 0:
            raise ValueError("Number of fuel rings must be >= 1.")
        self._num_fuel_rings = num_fuel_rings

        # Mass of fissionable matter in g / cm (i.e. per unit length)
        self._initial_fissionable_linear_mass = (
            fuel.fissionable_grams_per_cm3 * np.pi * self._fuel_radius**2.0
        )

        # Get gap related parameters
        if gap is None and gap_radius is not None:
            raise ValueError("Gap material is None but gap radius is defined.")
        elif gap is not None and gap_radius is None:
            raise ValueError("Gap material is defined but gap radius is None.")

        if gap_radius is not None and gap_radius <= fuel_radius:
            raise ValueError("Gap radius must be > fuel radius.")

        self._gap = copy.deepcopy(gap)
        self._gap_radius = gap_radius

        # Get cladding related parameters
        self._clad = copy.deepcopy(clad)

        if clad_radius <= 0.0 or clad_radius <= fuel_radius:
            raise ValueError("Clad radius must be > fuel radius.")
        elif gap_radius is not None and clad_radius <= gap_radius:
            raise ValueError("Clad radius must be > gap radius.")
        self._clad_radius = clad_radius

        # ======================================================================
        # DANCOFF CORRECTION CALCULATION DATA
        # ----------------------------------------------------------------------

        # Initialize empty list of Dancoff corrections for the fuel
        self._fuel_dancoff_corrections: List[float] = []

        # Initialize empty list of Dancoff corrections for the cladding
        self._clad_dancoff_corrections: List[float] = []

        # Initialize empty variables for Dancoff correction calculations.
        # These are all kept private.
        self._fuel_dancoff_xs: CrossSection = CrossSection(
            np.array([1.0e5]), np.array([1.0e5]), np.array([[0.0]]), "Fuel"
        )
        self._gap_dancoff_xs: Optional[CrossSection] = None
        if self.gap is not None:
            self._gap_dancoff_xs = CrossSection(
                np.array([self.gap.potential_xs]),
                np.array([self.gap.potential_xs]),
                np.array([[0.0]]),
                "Gap",
            )
        self._clad_dancoff_xs: CrossSection = CrossSection(
            np.array([self.clad.potential_xs]),
            np.array([self.clad.potential_xs]),
            np.array([[0.0]]),
            "Clad",
        )

        self._fuel_isolated_dancoff_fsr_ids = []
        self._gap_isolated_dancoff_fsr_ids = []
        self._clad_isolated_dancoff_fsr_ids = []
        self._mod_isolated_dancoff_fsr_ids = []

        self._fuel_full_dancoff_fsr_ids = []
        self._gap_full_dancoff_fsr_ids = []
        self._clad_full_dancoff_fsr_ids = []
        self._mod_full_dancoff_fsr_ids = []

        self._fuel_isolated_dancoff_fsr_inds = []
        self._gap_isolated_dancoff_fsr_inds = []
        self._clad_isolated_dancoff_fsr_inds = []
        self._mod_isolated_dancoff_fsr_inds = []

        self._fuel_full_dancoff_fsr_inds = []
        self._gap_full_dancoff_fsr_inds = []
        self._clad_full_dancoff_fsr_inds = []
        self._mod_full_dancoff_fsr_inds = []

        # ======================================================================
        # TRANSPORT CALCULATION DATA
        # ----------------------------------------------------------------------
        # Lists of the FSR IDs for each fuel ring, used to homogenize flux
        # spectra for depletion. These will be filled by make_moc_cell.
        self._fuel_ring_fsr_ids: List[List[int]] = []
        for r in range(self.num_fuel_rings):
            self._fuel_ring_fsr_ids.append([])
        self._gap_fsr_ids: List[int] = []
        self._clad_fsr_ids: List[int] = []
        self._mod_fsr_ids: List[int] = []

        self._fuel_ring_fsr_inds: List[List[int]] = []
        for r in range(self.num_fuel_rings):
            self._fuel_ring_fsr_inds.append([])
        self._gap_fsr_inds: List[int] = []
        self._clad_fsr_inds: List[int] = []
        self._mod_fsr_inds: List[int] = []

        # Create list of the different radii for fuel pellet
        self._fuel_radii = []
        if self.num_fuel_rings == 1:
            self._fuel_radii.append(self.fuel_radius)
        else:
            V = np.pi * self.fuel_radius * self.fuel_radius
            Vr = V / self.num_fuel_rings
            for ri in range(self.num_fuel_rings):
                Rin = 0.0
                if ri > 0:
                    Rin = self._fuel_radii[-1]
                Rout = np.sqrt((Vr + np.pi * Rin * Rin) / np.pi)
                if Rout > self.fuel_radius:
                    Rout = self.fuel_radius
                self._fuel_radii.append(Rout)

        # Initialize array of compositions for the fuel. This holds the
        # composition for each fuel ring and for each depletion step.
        self._fuel_ring_materials: List[List[Material]] = []
        for r in range(self.num_fuel_rings):
            # All rings initially start with the same composition
            self._fuel_ring_materials.append([copy.deepcopy(fuel)])

        # Initialize an array to hold the flux spectrum for each fuel ring.
        self._fuel_ring_flux_spectra: List[np.ndarray] = []
        for r in range(self.num_fuel_rings):
            # All rings initially start with empty flux spectrum list
            self._fuel_ring_flux_spectra.append(np.array([]))

        # Initialize an array to hold the depletion matrices for the previous and current steps.
        self._fuel_ring_prev_dep_mats: List[Optional[DepletionMatrix]] = []
        self._fuel_ring_current_dep_mats: List[Optional[DepletionMatrix]] = []
        for r in range(self.num_fuel_rings):
            # All rings initially start with empty matrix
            self._fuel_ring_prev_dep_mats.append(None)
            self._fuel_ring_current_dep_mats.append(None)

        # Holds all the CrossSection objects used for the real transport
        # calculation. These are NOT stored for each depletion step like with
        # the materials.
        self._fuel_ring_xs: List[CrossSection] = []
        self._gap_xs: Optional[CrossSection] = None
        self._clad_xs: Optional[CrossSection] = None

    @property
    def fuel_radius(self) -> float:
        return self._fuel_radius

    @property
    def num_fuel_rings(self) -> int:
        return self._num_fuel_rings

    @property
    def initial_fissionable_linear_mass(self) -> float:
        return self._initial_fissionable_linear_mass

    @property
    def fuel_ring_materials(self) -> List[List[Material]]:
        return self._fuel_ring_materials

    @property
    def fuel_ring_flux_spectra(self) -> List[List[Material]]:
        return self._fuel_ring_materials

    @property
    def fuel_dancoff_corrections(self) -> List[float]:
        return self._fuel_dancoff_corrections

    @property
    def gap(self) -> Optional[Material]:
        return self._gap

    @property
    def gap_radius(self) -> Optional[float]:
        return self._gap_radius

    @property
    def clad(self) -> Material:
        return self._clad

    @property
    def clad_radius(self) -> float:
        return self._clad_radius

    @property
    def clad_dancoff_corrections(self) -> List[float]:
        return self._clad_dancoff_corrections

    def _check_dx_dy(self, dx, dy, pintype):
        if pintype == PinCellType.Full:
            if dx < 2.0 * self.clad_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the diameter of the cladding."
                )
            if dy < 2.0 * self.clad_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the diameter of the cladding."
                )
        elif pintype in [PinCellType.XN, PinCellType.XP]:
            if dx < self.clad_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the radius of the cladding."
                )
            if dy < 2.0 * self.clad_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the diameter of the cladding."
                )
        elif pintype in [PinCellType.YN, PinCellType.YP]:
            if dy < self.clad_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the radius of the cladding."
                )
            if dx < 2.0 * self.clad_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the diameter of the cladding."
                )
        else:
            if dx < self.clad_radius:
                raise ValueError(
                    "The fuel pin cell x width must be > the radius of the cladding."
                )
            if dy < self.clad_radius:
                raise ValueError(
                    "The fuel pin cell y width must be > the radius of the cladding."
                )

    def load_nuclides(self, ndl: NDLibrary) -> None:
        """
        Loads all the nuclides for all current materials into the data library.

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library which should load the nuclides.
        """
        for ring_mats in self.fuel_ring_materials:
            ring_mats[-1].load_nuclides(ndl)

        if self.gap is not None:
            self.gap.load_nuclides(ndl)

        self.clad.load_nuclides(ndl)

    # ==========================================================================
    # Interrogation Methods
    def get_fuel_material(self, t: int, r: int = 0) -> Material:
        """
        Returns the Material object for a desired fuel ring at a desired
        depletion time step. Ring index 0 is at the center of the pin.

        Parameters
        ----------
        t : int
            Depletion time step index.
        r : int
            Ring index. Default is 0.

        Returns
        -------
        Material
            Material defining the temperature, density, and composition for the
            desired fuel ring and depletion time step.
        """
        if r >= self.num_fuel_rings:
            raise IndexError(f"Fuel ring index {r} is out of range.")

        if t >= len(self._fuel_ring_materials[r]):
            raise IndexError(f"Fuel time step index {t} is out of range.")

        return self._fuel_ring_materials[r][t]

    def get_average_fuel_nuclide_density(self, t: int, nuclide: str) -> float:
        """
        Computes the average density of a nuclide within the fuel pellet at a
        single depletion time step.

        Parameters
        ----------
        t : int
            Depletion time step index.
        nuclide : str
            Name of the nuclide.

        Returns
        -------
        float
            Average density of the nuclide at depletion time step t across the
            fuel pellet in units of atoms per barn-cm.
        """
        sum_density = 0.0

        for r in range(self.num_fuel_rings):
            mat = self.get_fuel_material(t, r)
            sum_density += mat.atom_density(nuclide)

        return sum_density / self.num_fuel_rings

    # ==========================================================================
    # Dancoff Correction Related Methods
    def set_xs_for_fuel_dancoff_calculation(self) -> None:
        """
        Sets the 1-group cross sections to calculate the fuel Dancoff correction.
        """
        self._fuel_dancoff_xs.set(
            CrossSection(
                np.array([1.0e5]), np.array([1.0e5]), np.array([[0.0]]), "Fuel"
            )
        )

        if self._gap_dancoff_xs is not None and self.gap is not None:
            self._gap_dancoff_xs.set(
                CrossSection(
                    np.array([self.gap.potential_xs]),
                    np.array([self.gap.potential_xs]),
                    np.array([[0.0]]),
                    "Gap",
                )
            )

        self._clad_dancoff_xs.set(
            CrossSection(
                np.array([self.clad.potential_xs]),
                np.array([self.clad.potential_xs]),
                np.array([[0.0]]),
                "Clad",
            )
        )

    def set_xs_for_clad_dancoff_calculation(self, ndl: NDLibrary) -> None:
        """
        Sets the 1-group cross sections to calculate the clad Dancoff correction.

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        # Create average fuel mixture
        fuel_mats = []
        fuel_vols = []
        for ring in self.fuel_ring_materials:
            fuel_mats.append(ring[-1])
            fuel_vols.append(1.0 / self.num_fuel_rings)
        avg_fuel: Material = mix_materials(
            fuel_mats, fuel_vols, MixingFraction.Volume, ndl
        )

        self._fuel_dancoff_xs.set(
            CrossSection(
                np.array([avg_fuel.potential_xs]),
                np.array([avg_fuel.potential_xs]),
                np.array([[0.0]]),
                "Fuel",
            )
        )

        if self._gap_dancoff_xs is not None and self.gap is not None:
            self._gap_dancoff_xs.set(
                CrossSection(
                    np.array([self.gap.potential_xs]),
                    np.array([self.gap.potential_xs]),
                    np.array([[0.0]]),
                    "Gap",
                )
            )

        self._clad_dancoff_xs.set(
            CrossSection(
                np.array([1.0e5]),
                np.array([1.0e5]),
                np.array([[0.0]]),
                "Clad",
            )
        )

    def make_dancoff_moc_cell(
        self,
        moderator_xs: CrossSection,
        dx: float,
        dy: float,
        pintype: PinCellType,
        isolated: bool,
    ) -> SimplePinCell:
        """
        Makes a simplified cell suitable for performing Dancoff correction
        calculations. The flat source region IDs are stored locally in the
        FuelPin object.

        Parameters
        ----------
        moderator_xs : CrossSection
            One group cross sections for the moderator. Total should equal
            absorption (i.e. no scattering) and should be equal to the
            macroscopic potential cross section.
        dx : float
            Width of the cell along x.
        dy : float
            Width of the cell along y.
        pintype : PinCellType
            How the pin cell should be split (along x, y, or only a quadrant).
        isolated : bool
            If True, the FSR IDs are stored for the isolated pin. Otherwise,
            they are stored for the full pin.

        Returns
        -------
        SimplifiedPinCell
            Pin cell object for MOC Dancoff correction calculation.
        """
        self._check_dx_dy(dx, dy, pintype)

        # First we create list of radii and materials
        radii = []
        xs = []

        radii.append(self.fuel_radius)
        xs.append(self._fuel_dancoff_xs)

        if self.gap is not None and self.gap_radius is not None:
            radii.append(self.gap_radius)
            xs.append(self._gap_dancoff_xs)

        radii.append(self.clad_radius)
        xs.append(self._clad_dancoff_xs)

        xs.append(moderator_xs)

        # Make the simple pin cell.
        cell = SimplePinCell(radii, xs, dx, dy, pintype)

        # Get the FSR IDs for the regions of interest
        cell_fsr_ids = list(cell.get_all_fsr_ids())
        cell_fsr_ids.sort()

        if isolated:
            self._fuel_isolated_dancoff_fsr_ids.append(cell_fsr_ids[0])
            if self.gap is None:
                self._clad_isolated_dancoff_fsr_ids.append(cell_fsr_ids[1])
                self._mod_isolated_dancoff_fsr_ids.append(cell_fsr_ids[2])
            else:
                self._gap_isolated_dancoff_fsr_ids.append(cell_fsr_ids[1])
                self._clad_isolated_dancoff_fsr_ids.append(cell_fsr_ids[2])
                self._mod_isolated_dancoff_fsr_ids.append(cell_fsr_ids[3])
        else:
            self._fuel_full_dancoff_fsr_ids.append(cell_fsr_ids[0])
            if self.gap is None:
                self._clad_full_dancoff_fsr_ids.append(cell_fsr_ids[1])
                self._mod_full_dancoff_fsr_ids.append(cell_fsr_ids[2])
            else:
                self._gap_full_dancoff_fsr_ids.append(cell_fsr_ids[1])
                self._clad_full_dancoff_fsr_ids.append(cell_fsr_ids[2])
                self._mod_full_dancoff_fsr_ids.append(cell_fsr_ids[3])

        return cell

    def populate_dancoff_fsr_indexes(
        self, isomoc: MOCDriver, fullmoc: MOCDriver
    ) -> None:
        """
        Obtains the flat source region indexes for all of the flat source
        regions used in the Dancoff correction calculations.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated pin.
        fullmoc : MOCDriver
            MOC simulation for the full geometry.
        """
        self._fuel_isolated_dancoff_fsr_inds = []
        self._gap_isolated_dancoff_fsr_inds = []
        self._clad_isolated_dancoff_fsr_inds = []
        self._mod_isolated_dancoff_fsr_inds = []

        self._fuel_full_dancoff_fsr_inds = []
        self._gap_full_dancoff_fsr_inds = []
        self._clad_full_dancoff_fsr_inds = []
        self._mod_full_dancoff_fsr_inds = []

        for id in self._fuel_isolated_dancoff_fsr_ids:
            self._fuel_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))
        for id in self._gap_isolated_dancoff_fsr_ids:
            self._gap_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))
        for id in self._clad_isolated_dancoff_fsr_ids:
            self._clad_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))
        for id in self._mod_isolated_dancoff_fsr_ids:
            self._mod_isolated_dancoff_fsr_inds.append(isomoc.get_fsr_indx(id, 0))

        for id in self._fuel_full_dancoff_fsr_ids:
            self._fuel_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))
        for id in self._gap_full_dancoff_fsr_ids:
            self._gap_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))
        for id in self._clad_full_dancoff_fsr_ids:
            self._clad_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))
        for id in self._mod_full_dancoff_fsr_ids:
            self._mod_full_dancoff_fsr_inds.append(fullmoc.get_fsr_indx(id, 0))

    def set_isolated_dancoff_fuel_sources(
        self, isomoc: MOCDriver, moderator: Material
    ) -> None:
        """
        Initializes the fixed sources for the isolated MOC calculation required
        in computing Dancoff corrections. Sources are set for a fuel Dancoff
        corrections calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        """
        # Fuel sources should all be zero !
        for ind in self._fuel_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, 0.0)

        # Gap sources should all be potential_xs
        if self.gap is not None:
            pot_xs = self.gap.potential_xs
            for ind in self._gap_isolated_dancoff_fsr_inds:
                isomoc.set_extern_src(ind, 0, pot_xs)

        # Clad sources should all be potential_xs
        pot_xs = self.clad.potential_xs
        for ind in self._clad_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

    def set_isolated_dancoff_clad_sources(
        self, isomoc: MOCDriver, moderator: Material, ndl: NDLibrary
    ) -> None:
        """
        Initializes the fixed sources for the isolated MOC calculation required
        in computing Dancoff corrections. Sources are set for a clad Dancoff
        correction calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        # Create average fuel mixture
        fuel_mats = []
        fuel_vols = []
        for ring in self.fuel_ring_materials:
            fuel_mats.append(ring[-1])
            fuel_vols.append(1.0 / self.num_fuel_rings)
        avg_fuel: Material = mix_materials(
            fuel_mats, fuel_vols, MixingFraction.Volume, ndl
        )

        # Fuel sources should all be potential_xs
        pot_xs = avg_fuel.potential_xs
        for ind in self._fuel_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

        # Gap sources should all be potential_xs
        if self.gap is not None:
            pot_xs = self.gap.potential_xs
            for ind in self._gap_isolated_dancoff_fsr_inds:
                isomoc.set_extern_src(ind, 0, pot_xs)

        # Clad sources should all be zero !
        for ind in self._clad_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, 0.0)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_isolated_dancoff_fsr_inds:
            isomoc.set_extern_src(ind, 0, pot_xs)

    def set_full_dancoff_fuel_sources(
        self, fullmoc: MOCDriver, moderator: Material
    ) -> None:
        """
        Initializes the fixed sources for the full MOC calculation required
        in computing Dancoff corrections. Sources are set for a fuel Dancoff
        correction calculation.

        Parameters
        ----------
        fullmoc : MOCDriver
            MOC simulation for the full geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        """
        # Fuel sources should all be zero !
        for ind in self._fuel_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, 0.0)

        # Gap sources should all be potential_xs
        if self.gap is not None:
            pot_xs = self.gap.potential_xs
            for ind in self._gap_full_dancoff_fsr_inds:
                fullmoc.set_extern_src(ind, 0, pot_xs)

        # Clad sources should all be potential_xs
        pot_xs = self.clad.potential_xs
        for ind in self._clad_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

    def set_full_dancoff_clad_sources(
        self, fullmoc: MOCDriver, moderator: Material, ndl: NDLibrary
    ) -> None:
        """
        Initializes the fixed sources for the full MOC calculation required
        in computing Dancoff corrections. Sources are set for a clad Dancoff
        correction calculation.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry.
        moderator : Material
            Material definition for the moderator, used to obtain the potential
            scattering cross section.
        ndl : NDLibrary
            Nuclear data library for obtaining potential scattering cross
            sections.
        """
        # Create average fuel mixture
        fuel_mats = []
        fuel_vols = []
        for ring in self.fuel_ring_materials:
            fuel_mats.append(ring[-1])
            fuel_vols.append(1.0 / self.num_fuel_rings)
        avg_fuel: Material = mix_materials(
            fuel_mats, fuel_vols, MixingFraction.Volume, ndl
        )

        # Fuel sources should all be potential_xs
        pot_xs = avg_fuel.potential_xs
        for ind in self._fuel_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

        # Gap sources should all be potential_xs
        if self.gap is not None:
            pot_xs = self.gap.potential_xs
            for ind in self._gap_full_dancoff_fsr_inds:
                fullmoc.set_extern_src(ind, 0, pot_xs)

        # Clad sources should all be zero !
        for ind in self._clad_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, 0.0)

        # Moderator sources should all be potential_xs
        pot_xs = moderator.potential_xs
        for ind in self._mod_full_dancoff_fsr_inds:
            fullmoc.set_extern_src(ind, 0, pot_xs)

    def compute_fuel_dancoff_correction(
        self, isomoc: MOCDriver, fullmoc: MOCDriver
    ) -> float:
        """
        Computes the Dancoff correction for the fuel region of the fuel pin.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry (previously solved).
        fullmoc : MOCDriver
            MOC simulation for the full geometry (previously solved).

        Returns
        -------
        float
            Dancoff correction for the fuel region.
        """
        iso_flux = isomoc.homogenize_flux_spectrum(
            self._fuel_isolated_dancoff_fsr_inds
        )[0]
        full_flux = fullmoc.homogenize_flux_spectrum(self._fuel_full_dancoff_fsr_inds)[
            0
        ]
        return (iso_flux - full_flux) / iso_flux

    def compute_clad_dancoff_correction(
        self, isomoc: MOCDriver, fullmoc: MOCDriver
    ) -> float:
        """
        Computes the Dancoff correction for the cladding region of the fuel pin.

        Parameters
        ----------
        isomoc : MOCDriver
            MOC simulation for the isolated geometry (previously solved).
        fullmoc : MOCDriver
            MOC simulation for the full geometry (previously solved).

        Returns
        -------
        float
            Dancoff correction for the cladding region.
        """
        iso_flux = isomoc.homogenize_flux_spectrum(
            self._clad_isolated_dancoff_fsr_inds
        )[0]
        full_flux = fullmoc.homogenize_flux_spectrum(self._clad_full_dancoff_fsr_inds)[
            0
        ]
        return (iso_flux - full_flux) / iso_flux

    def append_fuel_dancoff_correction(self, C) -> None:
        """
        Saves new Dancoff correction for the fuel that will be used for all
        subsequent cross section updates.

        Parameters
        ----------
        C : float
            New Dancoff correction.
        """
        if C < 0.0 or C > 1.0:
            raise ValueError(
                f"Dancoff correction must be in range [0, 1]. Was provided {C}."
            )
        self._fuel_dancoff_corrections.append(C)

    def append_clad_dancoff_correction(self, C) -> None:
        """
        Saves new Dancoff correction for the cladding that will be used for all
        subsequent cross section updates.

        Parameters
        ----------
        C : float
            New Dancoff correction.
        """
        if C < 0.0 or C > 1.0:
            raise ValueError(
                f"Dancoff correction must be in range [0, 1]. Was provided {C}."
            )
        self._clad_dancoff_corrections.append(C)

    # ==========================================================================
    # Transport Calculation Related Methods
    def set_fuel_xs_for_depletion_step(self, t: int, ndl: NDLibrary) -> None:
        """
        Constructs the CrossSection object for all fuel rings of the pin at the
        specified depletion step.

        Parameters
        ----------
        t : int
            Index for the depletion step.
        ndl : NDLibrary
            Nuclear data library to use for cross sections.
        """
        # Do the fuel cross sections
        if len(self._fuel_ring_xs) == 0:
            # Create initial CrossSection objects
            if self.num_fuel_rings == 1:
                # Compute escape xs
                Ee = 1.0 / (2.0 * self.fuel_radius)
                self._fuel_ring_xs.append(
                    self._fuel_ring_materials[0][t].carlvik_xs(
                        self._fuel_dancoff_corrections[t], Ee, ndl
                    )
                )
                if self._fuel_ring_xs[-1].name == "":
                    self._fuel_ring_xs[-1].name = "Fuel"
            else:
                # Do each ring
                for ri in range(self.num_fuel_rings):
                    Rin = 0.0
                    if ri > 0:
                        Rin = self._fuel_radii[ri - 1]
                    Rout = self._fuel_radii[ri]
                    self._fuel_ring_xs.append(
                        self._fuel_ring_materials[ri][t].ring_carlvik_xs(
                            self._fuel_dancoff_corrections[t],
                            self.fuel_radius,
                            Rin,
                            Rout,
                            ndl,
                        )
                    )
                    if self._fuel_ring_xs[-1].name == "":
                        self._fuel_ring_xs[-1].name = "Fuel"

        elif len(self._fuel_ring_xs) == self.num_fuel_rings:
            # Reset XS values. Cannot reassign or pointers will be broken !
            if self.num_fuel_rings == 1:
                # Compute escape xs
                Ee = 1.0 / (2.0 * self.fuel_radius)
                self._fuel_ring_xs[0].set(
                    self._fuel_ring_materials[0][t].carlvik_xs(
                        self._fuel_dancoff_corrections[t], Ee, ndl
                    )
                )
                if self._fuel_ring_xs[0].name == "":
                    self._fuel_ring_xs[0].name = "Fuel"
            else:
                # Do each ring
                for ri in range(self.num_fuel_rings):
                    Rin = 0.0
                    if ri > 0:
                        Rin = self._fuel_radii[ri - 1]
                    Rout = self._fuel_radii[ri]
                    self._fuel_ring_xs[ri].set(
                        self._fuel_ring_materials[ri][t].ring_carlvik_xs(
                            self._fuel_dancoff_corrections[t],
                            self.fuel_radius,
                            Rin,
                            Rout,
                            ndl,
                        )
                    )
                    if self._fuel_ring_xs[ri].name == "":
                        self._fuel_ring_xs[ri].name = "Fuel"
        else:
            raise RuntimeError(
                "Number of fuel cross sections does not agree with the number of fuel rings."
            )

    def set_gap_xs(self, ndl: NDLibrary) -> None:
        """
        Constructs the CrossSection object for the gap between the fuel pellet
        and the cladding of the pin.

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library to use for cross sections.
        """
        if self.gap is not None:
            if self._gap_xs is None:
                self._gap_xs = self.gap.dilution_xs([1.0e10] * self.gap.size, ndl)
            else:
                self._gap_xs.set(self.gap.dilution_xs([1.0e10] * self.gap.size, ndl))

            if self._gap_xs.name == "":
                self._gap_xs.name = "Gap"

    def set_clad_xs_for_depletion_step(self, t: int, ndl: NDLibrary) -> None:
        """
        Constructs the CrossSection object for the cladding of the pin at the
        specified depletion step. The depletion step only changes the Dancoff
        correction, not the cladding composition.

        Parameters
        ----------
        t : int
            Index for the depletion step.
        ndl : NDLibrary
            Nuclear data library to use for cross sections.
        """
        # Compute escape xs
        Ee = 0.0
        if self.gap_radius is not None:
            Ee = 1.0 / (2.0 * (self.clad_radius - self.gap_radius))
        else:
            Ee = 1.0 / (2.0 * (self.clad_radius - self.fuel_radius))

        # Get / set the xs
        if self._clad_xs is None:
            self._clad_xs = self.clad.roman_xs(
                self._clad_dancoff_corrections[t], Ee, ndl
            )
        else:
            self._clad_xs.set(
                self.clad.roman_xs(self._clad_dancoff_corrections[t], Ee, ndl)
            )

        if self._clad_xs.name == "":
            self._clad_xs.name = "Clad"

    def make_moc_cell(
        self,
        moderator_xs: CrossSection,
        dx: float,
        dy: float,
        pintype: PinCellType,
    ) -> PinCell:
        """
        Constructs the pin cell object used in for the global MOC simulation.

        Parameters
        ----------
        moderator_xs : CrossSection
            Cross sections to use for the moderator surrounding the fuel pin.
        dx : float
            Width of the cell along x.
        dy : float
            Width of the cell along y.
        pintype : PinCellType
            How the pin cell should be split (along x, y, or only a quadrant).

        Returns
        -------
        PinCell
            Pin cell suitable for the true MOC calculation.
        """
        if len(self._fuel_ring_xs) != self.num_fuel_rings:
            raise RuntimeError("Fuel cross sections have not yet been built.")
        if self.gap is not None and self._gap_xs is None:
            raise RuntimeError("Gap cross section has not yet been built.")
        if self._clad_xs is None:
            raise RuntimeError("Clad cross section has not yet been built.")
        self._check_dx_dy(dx, dy, pintype)

        # Initialize the radii and cross section lists with the fuel info
        radii = [r for r in self._fuel_radii]
        xss = [xs for xs in self._fuel_ring_xs]

        # Add the gap (if present)
        if self._gap_xs is not None:
            radii.append(self.gap_radius)
            xss.append(self._gap_xs)

        # Add cladding
        radii.append(self.clad_radius)
        xss.append(self._clad_xs)

        # Add another ring of moderator if possible
        if pintype == PinCellType.Full and min(dx, dy) > 2.0 * self.clad_radius:
            radii.append(0.5 * min(dx, dy))
            xss.append(moderator_xs)
        elif (
            pintype in [PinCellType.XN, PinCellType.XP]
            and dx > self.clad_radius
            and dy > 2.0 * self.clad_radius
        ):
            radii.append(min(dx, 0.5 * dy))
            xss.append(moderator_xs)
        elif (
            pintype in [PinCellType.YN, PinCellType.YP]
            and dy > self.clad_radius
            and dx > 2.0 * self.clad_radius
        ):
            radii.append(min(0.5 * dx, dy))
            xss.append(moderator_xs)
        elif dx > self.clad_radius and dy > self.clad_radius:
            radii.append(min(dx, dy))
            xss.append(moderator_xs)

        # Add moderator to the end of materials
        xss.append(moderator_xs)

        # Create the cell object
        cell = PinCell(radii, xss, dx, dy, pintype)

        # Get the FSR IDs for the regions of interest
        cell_fsr_ids = list(cell.get_all_fsr_ids())
        cell_fsr_ids.sort()

        # Number of angular divisions
        NA = 8
        if pintype in [PinCellType.XN, PinCellType.XP, PinCellType.YN, PinCellType.YP]:
            NA = 4
        elif pintype in [
            PinCellType.I,
            PinCellType.II,
            PinCellType.III,
            PinCellType.IV,
        ]:
            NA = 2

        I = 0  # Starting index for cell_fsr_inds
        # Go through all rings, and get FSR IDs
        for r in range(self.num_fuel_rings):
            for a in range(NA):
                self._fuel_ring_fsr_ids[r].append(cell_fsr_ids[I])
                I += 1

        # Get the FSRs for the gap, if present
        if self._gap_xs is not None:
            for a in range(NA):
                self._gap_fsr_ids.append(cell_fsr_ids[I])
                I += 1

        # Get the FSRs for the cladding
        for a in range(NA):
            self._clad_fsr_ids.append(cell_fsr_ids[I])
            I += 1

        # Everything else should be a moderator FSR
        self._mod_fsr_ids = list(cell_fsr_ids[I:])

        return cell

    def populate_fsr_indexes(self, moc: MOCDriver) -> None:
        """
        Obtains the flat source region indexes for all of the flat source
        regions used in the full MOC calculations.

        Parameters
        ----------
        moc : MOCDriver
            MOC simulation for the full calculations.
        """
        self._fuel_ring_fsr_inds: List[List[int]] = []
        for r in range(self.num_fuel_rings):
            self._fuel_ring_fsr_inds.append([])
        self._gap_fsr_inds: List[int] = []
        self._clad_fsr_inds: List[int] = []
        self._mod_fsr_inds: List[int] = []

        for r in range(self.num_fuel_rings):
            for id in self._fuel_ring_fsr_ids[r]:
                self._fuel_ring_fsr_inds[r].append(moc.get_fsr_indx(id, 0))
        for id in self._gap_fsr_ids:
            self._gap_fsr_inds.append(moc.get_fsr_indx(id, 0))
        for id in self._clad_fsr_ids:
            self._clad_fsr_inds.append(moc.get_fsr_indx(id, 0))
        for id in self._mod_fsr_ids:
            self._mod_fsr_inds.append(moc.get_fsr_indx(id, 0))

    def obtain_flux_spectra(self, moc: MOCDriver) -> None:
        """
        Computes the average flux spectrum for each fuel ring from the MOC
        simulation. Each ring's flux spectrum is volume averaged.

        Parameters
        ----------
        moc : MOCDriver
            MOC simulation for the full calculations.
        """
        for r in range(self.num_fuel_rings):
            self._fuel_ring_flux_spectra[r] = moc.homogenize_flux_spectrum(
                self._fuel_ring_fsr_inds[r]
            )

    def compute_pin_linear_power(self, ndl: NDLibrary):
        """
        Computes the linear power density of the fuel pin based on the current
        flux spectra, in units of w / cm. Does not consider the partial pin
        geometry at the assembly level (i.e. a half pin in a quarter assembly).

        Parameters
        ----------
        ndl : NDLibrary
            Nuclear data library for the fission energy release.

        Returns
        -------
        float
            Linear power density in w / cm.
        """
        power = 0.0
        A = np.pi * self.fuel_radius**2.0 / self.num_fuel_rings
        for r in range(self.num_fuel_rings):
            mat = self._fuel_ring_materials[r][-1]
            flux = self._fuel_ring_flux_spectra[r]
            power += A * mat.compute_fission_power_density(flux, ndl)
        # Convert from MeV/cm/s to J/cm/s = w/cm
        power *= 1.6021766339999e-13
        return power

    def normalize_flux_spectrum(self, f) -> None:
        """
        Applies a multiplicative factor to the flux spectra for the fuel rings.
        This permits normalizing the flux to a known assembly power.

        Parameters
        ----------
        f : float
            Normalization factor.
        """
        if f <= 0.0:
            raise ValueError("Normalization factor must be > 0.")

        for r in range(self.num_fuel_rings):
            self._fuel_ring_flux_spectra[r] *= f

    def predict_depletion(
        self,
        chain: DepletionChain,
        ndl: NDLibrary,
        dt: float,
        dtm1: Optional[float] = None,
    ) -> None:
        """
        Performs the predictor in the integration of the Bateman equation.
        If the argument for the previous time step is not provided, CE/LI will
        be used. Otherwise, CE/LI is used on the first depletion step, and
        LE/QI is used for all subsequent time steps. The predicted material
        compositions are appended to the materials lists.

        Paramters
        ---------

        chain : DepletionChain
            Depletion chain to use for radioactive decay and transmutation.
        ndl : NDLibrary
            Nuclear data library.
        dt : float
            Durration of the time step in seconds.
        dtm1 : float, optional
            Durration of the previous time step in seconds. Default is None.
        """
        if dt <= 0:
            raise ValueError("Predictor time step must be > 0.")

        # Do the prediction step for each fuel ring
        for r in range(self.num_fuel_rings):
            # Get the flux and initial material
            flux = self._fuel_ring_flux_spectra[r]
            mat = self._fuel_ring_materials[r][-1]  # Use last available mat !

            # Build depletion matrix for beginning of time step
            A0 = build_depletion_matrix(chain, mat, flux, ndl)

            # Save current matrix
            self._fuel_ring_current_dep_mats[r] = A0

            # At this point, we can clear the xs data from the last material as
            # depletion matrix is now built.
            mat.clear_all_micro_xs_data()

            # Initialize an array with the initial target number densities
            N = np.zeros(A0.size)
            nuclides = A0.nuclides
            for i, nuclide in enumerate(nuclides):
                N[i] = mat.atom_density(nuclide)

            if self._fuel_ring_prev_dep_mats[r] is None or dtm1 is None:
                # Use CE/LI
                A0 *= dt

                # Do the matrix exponential
                A0.exponential_product(N)

                # Undo multiplication by time step on the matrix
                A0 /= dt

            else:
                # Use LE/QI
                Am1 = self._fuel_ring_prev_dep_mats[r]

                F1 = (-dt / (12.0 * dtm1)) * Am1 + (
                    (6.0 * dtm1 + dt) / (12.0 * dtm1)
                ) * A0
                F1 *= dt

                F2 = (-5.0 * dt / (12.0 * dtm1)) * Am1 + (
                    (6.0 * dtm1 + 5.0 * dt) / (12.0 * dtm1)
                ) * A0
                F2 *= dt

                # Do the matrix exponentials
                F1.exponential_product(N)
                F2.exponential_product(N)

            # Now we can build a new material composition
            new_mat_comp = MaterialComposition()
            for i, nuclide in enumerate(nuclides):
                if N[i] > 0.0:
                    new_mat_comp.add_nuclide(nuclide, N[i])

            # Make the new material
            new_mat = Material(new_mat_comp, mat.temperature, ndl)
            self._fuel_ring_materials[r].append(new_mat)

    def correct_depletion(
        self,
        chain: DepletionChain,
        ndl: NDLibrary,
        dt: float,
        dtm1: Optional[float] = None,
    ) -> None:
        """
        Performs the corrector in the integration of the Bateman equation.
        If the argument for the previous time step is not provided, CE/LI will
        be used. Otherwise, CE/LI is used on the first depletion step, and
        LE/QI is used for all subsequent time steps. The corrected material
        compositions replace the ones where were appended in the corrector step.

        Parameters
        ----------
        chain : DepletionChain
            Depletion chain to use for radioactive decay and transmutation.
        ndl : NDLibrary
            Nuclear data library.
        dt : float
            Durration of the time step in seconds.
        dtm1 : float, optional
            Durration of the previous time step in seconds. Default is None.
        """
        if dt <= 0:
            raise ValueError("Corrector time step must be > 0.")

        # Do the prediction step for each fuel ring
        for r in range(self.num_fuel_rings):
            # Get the flux and initial material
            flux = self._fuel_ring_flux_spectra[r]
            mat_pred = self._fuel_ring_materials[r][-1]  # Use last available mat !

            # Get depletion matrix for beginning of time step
            A0 = self._fuel_ring_current_dep_mats[r]

            # Build depletion matrix and multiply by time step
            Ap1 = build_depletion_matrix(chain, mat_pred, flux, ndl)

            # Initialize an array with the initial target number densities
            mat_old = self._fuel_ring_materials[r][-2]  # Go 2 steps back !!
            N = np.zeros(Ap1.size)
            nuclides = Ap1.nuclides
            for i, nuclide in enumerate(nuclides):
                N[i] = mat_old.atom_density(nuclide)

            if self._fuel_ring_prev_dep_mats[r] is None or dtm1 is None:
                # Use CE/LI
                F1 = (5.0 * dt / 12.0) * A0 + (dt / 12.0) * Ap1
                F2 = (dt / 12.0) * A0 + (5.0 * dt / 12.0) * Ap1

                F1.exponential_product(N)
                F2.exponential_product(N)

            else:
                # Use LE/QI

                # Get previous depletion matrix
                Am1 = self._fuel_ring_prev_dep_mats[r]

                F3 = (
                    (-dt * dt / (12.0 * dtm1 * (dtm1 + dt))) * Am1
                    + (
                        (5.0 * dtm1 * dtm1 + 6.0 * dtm1 * dt + dt * dt)
                        / (12.0 * dtm1 * (dtm1 + dt))
                    )
                    * A0
                    + (dtm1 / (12.0 * (dtm1 + dt))) * Ap1
                )
                F3 *= dt

                F4 = (
                    (-dt * dt / (12.0 * dtm1 * (dtm1 + dt))) * Am1
                    + (
                        (dtm1 * dtm1 + 2.0 * dtm1 * dt + dt * dt)
                        / (12.0 * dtm1 * (dtm1 + dt))
                    )
                    * A0
                    + ((5.0 * dtm1 + 4.0 * dt) / (12.0 * (dtm1 + dt))) * Ap1
                )
                F4 *= dt

                F3.exponential_product(N)
                F4.exponential_product(N)

            # Now we can build a new material composition
            new_mat_comp = MaterialComposition()
            for i, nuclide in enumerate(nuclides):
                if N[i] > 0.0:
                    new_mat_comp.add_nuclide(nuclide, N[i])

            # Make the new material
            new_mat = Material(new_mat_comp, mat_pred.temperature, ndl)
            self._fuel_ring_materials[r][-1] = new_mat

            # Save the current matrix as previous matrix for next step !
            self._fuel_ring_prev_dep_mats[r] = A0
            self._fuel_ring_current_dep_mats[r] = None



# Analytical Benchmark Test Set for Criticality Code and Verification
# Problem No.: 36 [Ua-1-1-CY]
# Exact keff = 1.

from scarabee import *
import numpy as np
import matplotlib.pyplot as plt


Et = np.array([0.32640])
Ea = np.array([0.32640-0.248064])
Es = np.array([[[0.248064]],
               [[0.042432]]])
Ef = np.array([0.065280])
vEf = np.array([2.70 * 0.065280])
chi = np.array([1.])
dtr = np.array([0.042432])
fuel = CrossSection(Et, dtr, Ea, Es, Ef, vEf, chi)

# fictitious material
Ea = np.array([1E-4])
Et = Ea
Es = np.array([[[0.]],
               [[0.]]])
Ef = np.array([0.])
vEf = np.array([0.])
chi = np.array([0.])
dtr = np.array([0.])
fictitious_mat = CrossSection(Et, dtr, Ea, Es, Ef, vEf, chi)

# Creating Cell

# max length in x and y
radius = 5.514296811
length_x = 2 * radius
length_y = 2 * radius
# number of elements in x and y
Nx = 100
Ny = 100
# delta x and y for element
dx = length_x / Nx
dy = length_y / Ny
# low points
x0 = -length_x * 0.5
y0 = -length_y * 0.5

# fuel-cell
F = EmptyCell(fuel, dx, dy)
# ficititious material cell
N = EmptyCell(fictitious_mat, dx, dy)

# creating the cylinder with small cubes
c2d_dx = [dx] * Nx
c2d_dy = [dy] * Ny

root_univ = Cartesian2D(c2d_dx, c2d_dy)

lattice = []
for j in range(0, Ny):
    # mid-y in the element
    y = y0 + (j + 0.5) * dy

    for i in range(0, Nx):
        # mid-x in the element
        x = x0 + (i + 0.5) * dx
        # check, if the mid point is inside the cylinder
        if (np.sqrt(x*x + y*y) < radius):
            lattice.append(F)
        else:
            lattice.append(N)
            
root_univ.set_tiles(lattice)

moc = MOCDriver(root_univ, anisotropic = True)
moc.x_min_bc = BoundaryCondition.Vacuum
moc.x_max_bc = BoundaryCondition.Vacuum
moc.y_min_bc = BoundaryCondition.Vacuum
moc.y_max_bc = BoundaryCondition.Vacuum
moc.generate_tracks(128, 0.01, YamamotoTabuchi6())
moc.flux_tolerance = 1.E-5

# moc.plot()
moc.solve()

#flux, x, y = moc.rasterize_flux(100, 100)





from .._scarabee import (
    CrossSection,
    EmptyCell,
    Cartesian2D,
)
from typing import List, Tuple


def _ensleeve_quarter(
    pin_geom: Cartesian2D, pitch: float, gap_width: float, gap_xs: CrossSection
) -> Tuple[Cartesian2D, List[int]]:
    #  ---|
    #   PG|

    # First make the gap on the right hand side going from bottom to top
    LY = pin_geom.dy + gap_width
    dy = []
    right_gap_tiles = []
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        right_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    right_gap_tiles.reverse()
    right_gap = Cartesian2D([gap_width], dy)
    right_gap.set_tiles(right_gap_tiles)

    # Make top gap
    LX = pin_geom.dx
    dx = []
    top_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        top_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    top_gap = Cartesian2D(dx, [gap_width])
    top_gap.set_tiles(top_gap_tiles)

    # Make the temp geometry which is the pin_geom with the top gap
    temp_geom = Cartesian2D([pin_geom.dx], [pin_geom.dy, gap_width])
    temp_geom.set_tiles([top_gap, pin_geom])

    # Make the sleeved geometry
    sleeved_geom = Cartesian2D([pin_geom.dx, gap_width], [pin_geom.dy + gap_width])
    sleeved_geom.set_tiles([temp_geom, right_gap])

    # Get the FSR IDs for all the gap cells
    gap_fsr_ids = []
    for cell in right_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in top_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())

    return sleeved_geom, gap_fsr_ids


def _ensleeve_half_top(
    pin_geom: Cartesian2D, pitch: float, gap_width: float, gap_xs: CrossSection
) -> Tuple[Cartesian2D, List[int]]:
    #  |--|
    #  |PG|

    # First make the gap on the right hand side going from bottom to top
    LY = pin_geom.dy + gap_width
    dy = []
    right_gap_tiles = []
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        right_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    right_gap_tiles.reverse()
    right_gap = Cartesian2D([gap_width], dy)
    right_gap.set_tiles(right_gap_tiles)

    # Make the gap on the left hand side going from bottom to top
    LY = pin_geom.dy + gap_width
    dy = []
    left_gap_tiles = []
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        left_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    left_gap_tiles.reverse()
    left_gap = Cartesian2D([gap_width], dy)
    left_gap.set_tiles(left_gap_tiles)

    # Make top gap
    LX = pin_geom.dx
    dx = []
    top_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        top_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    top_gap = Cartesian2D(dx, [gap_width])
    top_gap.set_tiles(top_gap_tiles)

    # Make the temp geometry which is the pin_geom with the top gap
    temp_geom = Cartesian2D([pin_geom.dx], [pin_geom.dy, gap_width])
    temp_geom.set_tiles([top_gap, pin_geom])

    # Make the sleeved geometry
    sleeved_geom = Cartesian2D(
        [gap_width, pin_geom.dx, gap_width], [pin_geom.dy + gap_width]
    )
    sleeved_geom.set_tiles([left_gap, temp_geom, right_gap])

    # Get the FSR IDs for all the gap cells
    gap_fsr_ids = []
    for cell in left_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in right_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in top_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())

    return sleeved_geom, gap_fsr_ids


def _ensleeve_half_right(
    pin_geom: Cartesian2D, pitch: float, gap_width: float, gap_xs: CrossSection
) -> Tuple[Cartesian2D, List[int]]:
    #  --|
    #  PG|
    #  --|

    # First make the gap on the right hand side going from bottom to top
    LY = pin_geom.dy + 2.0 * gap_width
    extra_y = 0.5 * (LY - 0.25 * pitch * round(LY / (0.25 * pitch)))
    dy = [0.25 * pitch + extra_y]
    right_gap_tiles = [EmptyCell(gap_xs, gap_width, 0.25 * pitch + extra_y)]
    LY -= 0.25 * pitch + extra_y
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        right_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    right_gap_tiles.reverse()
    right_gap = Cartesian2D([gap_width], dy)
    right_gap.set_tiles(right_gap_tiles)

    # Make top gap
    LX = pin_geom.dx
    dx = []
    top_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        top_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    top_gap = Cartesian2D(dx, [gap_width])
    top_gap.set_tiles(top_gap_tiles)

    # Make bottom gap
    LX = pin_geom.dx
    dx = []
    bot_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        bot_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    bot_gap = Cartesian2D(dx, [gap_width])
    bot_gap.set_tiles(bot_gap_tiles)

    # Make the temp geometry which is the pin_geom with the top and bottom gap
    temp_geom = Cartesian2D([pin_geom.dx], [gap_width, pin_geom.dy, gap_width])
    temp_geom.set_tiles([top_gap, pin_geom, bot_gap])

    # Make the sleeved geometry
    sleeved_geom = Cartesian2D(
        [pin_geom.dx, gap_width], [pin_geom.dy + 2.0 * gap_width]
    )
    sleeved_geom.set_tiles([temp_geom, right_gap])

    # Get the FSR IDs for all the gap cells
    gap_fsr_ids = []
    for cell in right_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in top_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in bot_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())

    return sleeved_geom, gap_fsr_ids


def _ensleeve_full(
    pin_geom: Cartesian2D, pitch: float, gap_width: float, gap_xs: CrossSection
) -> Tuple[Cartesian2D, List[int]]:
    #  |--|
    #  |PG|
    #  |--|

    # First make the gap on the right hand side going from bottom to top
    LY = pin_geom.dy + 2.0 * gap_width
    extra_y = 0.5 * (LY - 0.25 * pitch * round(LY / (0.25 * pitch)))
    dy = [0.25 * pitch + extra_y]
    right_gap_tiles = [EmptyCell(gap_xs, gap_width, 0.25 * pitch + extra_y)]
    LY -= 0.25 * pitch + extra_y
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        right_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    right_gap_tiles.reverse()
    right_gap = Cartesian2D([gap_width], dy)
    right_gap.set_tiles(right_gap_tiles)

    # Make the gap on the left hand side going from bottom to top
    LY = pin_geom.dy + 2.0 * gap_width
    LY = pin_geom.dy + 2.0 * gap_width
    extra_y = 0.5 * (LY - 0.25 * pitch * round(LY / (0.25 * pitch)))
    dy = [0.25 * pitch + extra_y]
    left_gap_tiles = [EmptyCell(gap_xs, gap_width, 0.25 * pitch + extra_y)]
    LY -= 0.25 * pitch + extra_y
    while LY > 0.0:
        ly = 0.25 * pitch
        break_after = False
        if LY > 0.25 * pitch and LY < 0.5 * pitch:
            ly = LY
            break_after = True
        LY -= ly
        dy.append(ly)
        left_gap_tiles.append(EmptyCell(gap_xs, gap_width, ly))
        if break_after:
            break
    left_gap_tiles.reverse()
    left_gap = Cartesian2D([gap_width], dy)
    left_gap.set_tiles(left_gap_tiles)

    # Make top gap
    LX = pin_geom.dx
    dx = []
    top_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        top_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    top_gap = Cartesian2D(dx, [gap_width])
    top_gap.set_tiles(top_gap_tiles)

    # Make bottom gap
    LX = pin_geom.dx
    dx = []
    bot_gap_tiles = []
    while LX > 0.0:
        lx = 0.25 * pitch
        break_after = False
        if LX > 0.25 * pitch and LX < 0.5 * pitch:
            lx = LX
            break_after = True
        LX -= lx
        dx.append(lx)
        bot_gap_tiles.append(EmptyCell(gap_xs, lx, gap_width))
        if break_after:
            break
    bot_gap = Cartesian2D(dx, [gap_width])
    bot_gap.set_tiles(bot_gap_tiles)

    # Make the temp geometry which is the pin_geom with the top gap
    temp_geom = Cartesian2D([pin_geom.dx], [gap_width, pin_geom.dy, gap_width])
    temp_geom.set_tiles([top_gap, pin_geom, bot_gap])

    # Make the sleeved geometry
    sleeved_geom = Cartesian2D(
        [gap_width, pin_geom.dx, gap_width], [pin_geom.dy + 2.0 * gap_width]
    )
    sleeved_geom.set_tiles([left_gap, temp_geom, right_gap])

    # Get the FSR IDs for all the gap cells
    gap_fsr_ids = []
    for cell in left_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in right_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in top_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())
    for cell in bot_gap_tiles:
        gap_fsr_ids.append(cell.get_all_fsr_ids())

    return sleeved_geom, gap_fsr_ids
